#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Addons and additions to the Zope site concept.

Integration between the ZCA ``site`` system, configured site
policies, and the persistent database.

In the Zope world, sites are objects that can express configuration by
holding onto an instance of IComponents known as its *site manager*.
Typically they are arranged in a tree, with the global site at the
root of the tree. Site managers inherit configuration from their
parents (bases, which may or may not be their ``__parent__``). Often,
they are persistent and part of the traversal tree. One site is the
current site and the ZCA functions (e.g.,
:meth:`.IComponentArchitecture.queryUtility`) apply to that site.

Our application has one main persistent site, the application site,
containing persistent utilities (such as the application object). This
site, or a descendent of it, must always be the current site when
executing application code.

In our application, we also have the concept of site policies,
something that is applied based on virtual hosting. A site policy is
also an ``IComponents``, registered in the global site as a utility
named for the hostname to which it should apply (e.g.,
``domain.example.com``). Historically, these are not necessarily
persistent and part of the traversal tree. Now, we expect them to be
persistent, though they are still not part of the traversal tree.

Thus there are two things to accomplish: make the application site the
current site, and also construct a site that descends from that site
and contains any applicable policies that are globally configured.

* See :mod:`nti.site.interfaces` for descriptions of the concepts this
  package uses.
* See :mod:`nti.site.hostpolicy` for working with registered and persistent
  site configurations.
* See :mod:`nti.site.site` for the getting the correctly configured site.
* See :mod:`nti.site.subscribers` for setting up the correct site during traversal.

"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

from nti.site.hostpolicy import get_host_site
from nti.site.hostpolicy import get_all_host_sites

from nti.site.utils import registerUtility
from nti.site.utils import unregisterUtility



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Subscribers for traversal events to ensure that the the proper
site is installed, including respecting global registered components.

.. caution:: :func:`zope.site.site.threadSiteSubscriber` also exists and is
   configured by some packages like :mod:`zope.app.publication`. Care must be used to
   ensure that it is *not* configured in place of our own :func:`threadSiteSubscriber`.
"""

# turn off warning for not calling superclass, calling indirect superclass and
# accessing protected methods. we're deliberately doing both
# pylint: disable=W0233,W0231,W0212

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

from zope import component

from zope.component.hooks import getSite
from zope.component.hooks import setSite

from zope.component.interfaces import ISite

from zope.location.interfaces import LocationError

from zope.proxy import ProxyBase
from zope.proxy import non_overridable

from zope.site.interfaces import IRootFolder
from zope.site.interfaces import INewLocalSite

from zope.traversing.interfaces import IBeforeTraverseEvent

from nti.site.interfaces import IHostPolicyFolder
from nti.site.interfaces import IMainApplicationFolder

from nti.site.transient import BasedSiteManager

class _ProxyTraversedSite(ProxyBase):
    """
    We need to be able to control the site manager used
    by sites we traverse to in order to ensure that host
    configuration is at the right place in the resolution order.
    But a site can be literally any type of object. So we fake out the
    siteManager methods but proxy everything else.
    """

    def __new__(cls, base, site_manager):
        return ProxyBase.__new__(cls, base)

    def __init__(self, base, site_manager):
        ProxyBase.__init__(self, base)
        self.__site_manager = site_manager

    @non_overridable
    def getSiteManager(self):
        return self.__site_manager

    @non_overridable
    def setSiteManager(self, new_man):
        raise ValueError("Cannot set site manager on proxy")

@component.adapter(ISite, IBeforeTraverseEvent)
def threadSiteSubscriber(new_site, event):
    """
    Set the current ``zope.component.hooks`` site to
    the ``new_site`` object found during traversal,
    being careful to maintain any previously installed host (site-name)
    configurations as lower priority than the new site.

    Sites encountered during traversal are expected to have the
    main application site (e.g., ``nti.dataserver``) in their base chain
    so we have access to its configuration and persistent utilities.
    This implies that sites encountered during traversal are either
    synthetic (generated by a traversal adapter to use some particular
    ``IComponents``)  or themselves persistent.

    Because of this, when we encounter the root or dataserver folders
    as sites, we take no action.

    We expect that something else takes care of clearing the site.
    """

    if IMainApplicationFolder.providedBy(new_site) or IRootFolder.providedBy(new_site):
        # TODO: Since we get these events, we could
        # actually replace nti.appserver.tweens.zope_site_tween
        # with this. That's probably the longterm answer.
        return

    current_site = getSite()
    if current_site is None:
        # Nothing to do
        setSite(new_site)
    elif current_site is new_site:
        # This is typically the case when we traverse directly
        # into utilities registered with the site, for example
        #   /dataserver2/++etc++hostsites/janux.ou.edu/++etc++site/SOMEUTILITY/...
        # with the current host being janux.ou.edu.
        pass
    elif IHostPolicyFolder.providedBy(current_site) and \
         IHostPolicyFolder.providedBy(new_site):
        # This is typically the case when we traverse directly
        # into utilities registered with the site, for example
        #   /dataserver2/++etc++hostsites/janux.ou.edu/++etc++site/SOMEUTILITY/...
        # with the current host NOT being janux.ou.edu.
        # We do not want to switch host configurations here, but we do
        # want to allow traversal, so we take no action.
        # TODO: We might want to only allow this if there is some
        # inheritance relationship between the two sites?
        pass
    elif hasattr(current_site.getSiteManager(), 'host_components'):
        # A site synthesized by get_site_for_site_names
        # OR one previously synthesized by this function. In either case,
        # we always want to proxy, putting the preserved host components
        # at the end of the new proxy RO.
        host_components = current_site.getSiteManager().host_components
        # We need to keep host_components in the bases
        # for the new site. Where to put it is tricky
        # if we want to support multiple layers of overriding
        # of host registrations. Fortunately, the zope.interface.ro
        # machinery does exactly the right thing if we tack host
        # components (which are probably not already in the list)
        # on to the end. If they are in the list already, they
        # stay where they were.
        new_bases = new_site.getSiteManager().__bases__ + (host_components,)
        # TODO: We don't need to proxy the site manager, right?
        # it's almost never special by itself...
        new_site_manager = BasedSiteManager(new_site.__parent__,
                                            new_site.__name__,
                                            new_bases)
        new_site_manager.host_components = host_components
        new_fake_site = _ProxyTraversedSite(new_site,
                                            new_site_manager)

        setSite(new_fake_site)
    else:
        # Cancel traversal using a LocationError. This typically
        # will get surfaced as a 404.
        raise LocationError("Unknown kind of site", new_site, current_site)

@component.adapter(INewLocalSite)
def new_local_site_dispatcher(event):
    """
    Dispatches just like an object event,
    that way we can do things based on the type of the
    site manager.

    Note that if the containing ISite is (re)moved, an
    ObjectEvent will be fired for (sitemanager, site-event);
    that is, you subscribe to the site manager and the object moved
    event, but the event will have the ISite as the object property.
    """
    component.handle(event.manager, event)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904

from hamcrest import is_
from hamcrest import is_not
from hamcrest import none
from hamcrest import raises
from hamcrest import calling
from hamcrest import assert_that
does_not = is_not

from nti.testing.base import AbstractTestBase
import fudge

from zope.interface import implementer

from zope.component import globalSiteManager as BASE
from zope.component.hooks import getSite

from z3c.baseregistry.baseregistry import BaseComponents

from nti.site.transient import HostSiteManager as HSM
from nti.site.transient import TrivialSite

from zope.site.interfaces import IRootFolder
from ..interfaces import IMainApplicationFolder

from ..subscribers import threadSiteSubscriber

class TestSubscriber(AbstractTestBase):

    def test_root_folder_not_set(self):
        @implementer(IRootFolder)
        class Root(object):
            pass

        threadSiteSubscriber(Root(), None)

        assert_that(getSite(), is_(none()))

    def test_main_folder_not_set(self):
        @implementer(IMainApplicationFolder)
        class Main(object):
            pass

        threadSiteSubscriber(Main(), None)

        assert_that(getSite(), is_(none()))

    @fudge.patch('nti.site.subscribers.getSite')
    def test_not_replace_same(self, fake_get):
        fake_get.is_callable().returns(self)

        # If we actually tried to set TestSubscriber as the site,
        # we'd blow up
        threadSiteSubscriber(self, None)

    @fudge.patch('nti.site.subscribers.getSite')
    def test_returns_unsettable_proxy(self, fake_get):
        pers_comps = BaseComponents(BASE, 'persistent', (BASE,))
        host_comps = BaseComponents(BASE, 'example.com', (BASE,))
        host_sm = HSM('example.com', 'siteman', host_comps, pers_comps)

        fake_get.is_callable().returns(TrivialSite(host_sm))

        host_comps2 = BaseComponents(BASE, 'other.com', (BASE,))
        host_sm2 = HSM('other.com', 'siteman', host_comps2, pers_comps)

        threadSiteSubscriber(TrivialSite(host_sm2), None)

        # Existing host comps are preserved
        assert_that(getSite().getSiteManager().host_components, is_(host_comps))

        # and it's proxied and we can't change it
        assert_that(calling(getSite().setSiteManager).with_args(None),
                    raises(ValueError, "Cannot set site manager on proxy"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""


.. $Id$
"""

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

import unittest
from hamcrest import assert_that
from hamcrest import is_

from hamcrest import has_length


class TestUtils(unittest.TestCase):

    def test_event_unregister_warning(self):

        import warnings
        from ..utils import unregisterUtility
        from zope.component import getGlobalSiteManager
        from zope.interface import Interface
        class IFoo(Interface):
            pass

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")

            gsm = getGlobalSiteManager()
            unregisterUtility(gsm, self, provided=IFoo, event=True)

            assert_that(w, has_length(1))
            assert_that(w[-1].category(), is_(FutureWarning))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""


.. $Id$
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

#disable: accessing protected members, too many methods
#pylint: disable=W0212,R0904

import unittest
from hamcrest import assert_that
from hamcrest import is_
from hamcrest import none
from hamcrest import is_not
from hamcrest import is_in

from nti.testing import base
from nti.testing.matchers import provides


from zope.interface.registry import UtilityRegistration
from zope.interface.interfaces import Registered

from zope.interface import Interface
from zope import interface

from zope.component.interfaces import ISite
from ..site import BTreeLocalSiteManager as LocalSiteManager
from zope.site import SiteManagerContainer
from zope.container.contained import Contained
from zope.location.interfaces import ILocationInfo

from ..localutility import install_utility_on_registration
from ..localutility import uninstall_utility_on_unregistration
from ..localutility import queryNextUtility

class TestInstallUninstall(base.AbstractTestBase):


    def test_install_uninstall(self):
        lsm = LocalSiteManager(None, None)
        registration = UtilityRegistration(lsm, None, None, None, None)
        event = Registered(registration)

        class IFoo(Interface):
            pass

        # If we're not contained, we get proxied!
        class Foo(Contained):
            pass

        install_utility_on_registration(Foo(), 'name', IFoo, event)

        assert_that(lsm['name'], is_(Foo))
        assert_that(lsm.getUtility(IFoo), is_(Foo))

        queryNextUtility(lsm['name'], IFoo)

        uninstall_utility_on_unregistration('name', IFoo, event)

        assert_that('name', is_not(is_in(lsm)))
        assert_that(lsm.queryUtility(IFoo), is_(none()))

class TestQueryNext(base.AbstractTestBase):

    def test_query(self):
        from zope.site import SiteManagerAdapter
        from zope import component
        component.provideAdapter(SiteManagerAdapter)
        top_site = SiteManagerContainer()
        top_sm = LocalSiteManager(top_site)
        top_site.setSiteManager(top_sm)
        assert_that(top_sm.__parent__, is_(top_site))
        assert_that(top_site, provides(ISite))
        interface.alsoProvides(top_site, ILocationInfo)

        child_site = SiteManagerContainer()
        child_site.__parent__ = top_site
        child_site.getParent = lambda: child_site.__parent__
        interface.alsoProvides(child_site, ILocationInfo)
        child_sm = LocalSiteManager(child_site)
        child_site.setSiteManager(child_sm)

        assert_that(child_sm.__bases__, is_((top_sm,)))


        class IFoo(Interface):
            pass

        @interface.implementer(IFoo)
        class Foo(Contained):
            pass

        child_foo = Foo()
        top_foo = Foo()

        child_foo.__parent__ = child_site
        top_foo.__parent__ = top_site

        top_sm.registerUtility(top_foo, IFoo)
        child_sm.registerUtility(child_foo, IFoo)

        child_foo.__conform__ = lambda self, _:  child_sm
        from zope import component
        component.getSiteManager(child_foo)


        x = queryNextUtility(child_foo, IFoo)
        assert_that(x, is_(top_foo))

        class IBaz(Interface): pass

        x = queryNextUtility(child_foo, IBaz)
        assert_that(x, is_(none()))

        x = queryNextUtility(top_foo, IFoo)
        assert_that(x, is_(none()))

        x = queryNextUtility(component.getGlobalSiteManager(), IFoo)
        assert_that(x, is_(none()))

        global_foo = Foo()
        component.provideUtility(global_foo, IFoo)
        x = queryNextUtility(component.getGlobalSiteManager(), IFoo)
        assert_that(x, is_(global_foo))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Support for host policies.

This contains the main unique functionality for this package.

.. $Id$
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

from six import string_types

from zope import lifecycleevent
from zope import component
from zope import interface

from zope.component.hooks import site as current_site
from zope.component.interfaces import IComponents
from zope.component.interfaces import ISite

from zope.interface import ro
from zope.interface.interfaces import ComponentLookupError

from zope.traversing.interfaces import IEtcNamespace


from zope.site.folder import Folder
from zope.site.folder import rootFolder

from .folder import HostPolicyFolder
from .folder import HostPolicySiteManager
from .folder import HostSitesFolder
from .interfaces import IMainApplicationFolder
from .site import BTreeLocalSiteManager


def synchronize_host_policies():
    """
    Called within a transaction with a site being the current application
    site, find any :mod:`z3c.baseregistry` components that
    should be persistent sites, and register them in the database.

    As a prerequisite, :func:`install_sites_folder` must have been done, and
    we must be in that site.
    """

    # TODO: We will ultimately need to deal with removing and renaming
    # of these

    # Resolution order: The actual ISite __parent__ order is not
    # important, so we can keep them flat to mirror the GSM IComponents
    # registrations. What matters is the __bases__ of the site managers.
    # Now, if the global IComponents are themselves flat, then it doesn't matter;
    # however, if you have a hierarchy (and we do) then it matters critically, because
    # we need to pick up persistent utilities from these objects, as well as
    # the global components, in the right order. For example, if we have this
    # global hierarchy:
    #  GSM
    #  \
    #   S1
    #   \
    #     S2
    # and in the database we have the nti.dataserver and root persistent site managers,
    # then when we create the persistent sites for S1 and S2 (PS1 and PS2) we want
    # the resolution order to be:
    #   PS2 -> S2 -> PS1 -> S1 -> DS -> Root -> GSM
    # That is, we need to get the persistent components mixed in between the
    # global components.
    # Fortunately this is very easy to achieve. The code in zope.interface.ro handles
    # this.
    # We just need to ensure:
    #   PS1.__bases__ = (S1, DS)
    #   PS2.__bases__ = (S2, PS1)

    sites = component.getUtility(IEtcNamespace, name='hostsites')
    ds_folder = sites.__parent__
    assert IMainApplicationFolder.providedBy(ds_folder)

    ds_site_manager = ds_folder.getSiteManager()

    # Ok, find everything that is globally registered
    global_sm = component.getGlobalSiteManager()
    all_global_named_utilities = list(global_sm.getUtilitiesFor(IComponents))
    for name, comp in all_global_named_utilities:
        # The sites must be registered the same as their internal name
        assert name == comp.__name__
    all_global_utilities = [x[1] for x in all_global_named_utilities]

    # Now, get the resolution order of each site; this is an easy way
    # to do a kind of topological sort.
    site_ros = [ro.ro(x) for x in all_global_utilities]

    # Next, start creating persistent sites in the database, walking from the top
    # of the resolution order (the end of the list)
    # towards the root; the first one we put in the DB gets the DS as its
    # base, otherwise it gets the previous one we put in.

    for site_ro in site_ros:
        site_ro = reversed(site_ro)

        secondary_comps = ds_site_manager
        for comps in site_ro:
            name = comps.__name__
            logger.debug("Checking host policy for site %s", name)
            if name.endswith('base') or name.startswith('base'):
                # The GSM or the base global objects
                # TODO: better way to do this...marker interface?
                continue # pragma: no cover
            if name in sites:
                logger.debug("Host policy for %s already in place", name)
                # Ok, we've already put one in for this level.
                # We need to make it our next choice going forward
                secondary_comps = sites[name].getSiteManager()
            else:
                # Great, create the site
                logger.info("Installing site policy %s", name)

                site = HostPolicyFolder()
                # should fire object created event
                sites[name] = site

                site_policy = HostPolicySiteManager(site)
                site_policy.__bases__ = (comps, secondary_comps)
                # should fire INewLocalSite
                site.setSiteManager(site_policy)
                secondary_comps = site_policy


def install_sites_folder(server_folder):
    """
    Given a :class:`.IMainApplicationFolder` that has a site manager,
    install a host sites folder.
    """
    sites = HostSitesFolder()
    str(sites) # coverage
    repr(sites) # coverage
    server_folder['++etc++hostsites'] = sites
    lsm = server_folder.getSiteManager()
    lsm.registerUtility(sites, provided=IEtcNamespace, name='hostsites')
    # synchronize_host_policies()

def install_main_application_and_sites(conn,
                                       root_name=u'Application',
                                       root_alias=u'nti.dataserver_root',
                                       main_name=u'dataserver2',
                                       main_alias=u'nti.dataserver',
                                       main_factory=Folder,
                                       main_setup=None):
    """
    Install the main application and site folder structure into ZODB.

    When this completes, the ZODB root object will have a :class:`.IRootFolder`
    object at "Application" (and optionally at *root_alias*). This will
    have a site manager.

    The root folder in turn will have a
    :class:`.IMainApplicationFolder` child named *main_name* (and
    optionally at *main_alias*). It will have a site manager, and in
    this site manager will be the "++etc++hostsites" object used to
    contain host site folders.

    :param conn: The open ZODB connection.
    :keyword str root_name: The main name of the root folder. This generally should
      be left as "Application" as that's what many Zope 3 components expect.
    :keyword main_factory: The factory that will be used for the :class:`.IMainApplicationFolder`.
      If it produces an object that doesn't implement this interface, it will still
      be marked as doing so.
    :keyword callable main_setup: If given, a callable that will accept the main
      application folder object and perform further setup on it. This will be called
      *before* any lifecycle events are generated. This is a good time to install additional
      utilities, such as :class:`IIntId` utilities.
    """
    root = conn.root()

    # The root folder
    root_folder = rootFolder()
    # NOTE that the root_folder doesn't get a __name__!
    conn.add(root_folder)  # Ensure we have a connection so we can become KeyRefs
    assert root_folder._p_jar is conn

    # The root is generally presumed to be an ISite, so make it so
    root_sm = BTreeLocalSiteManager(root_folder)  # site is IRoot, so __base__ is the GSM
    assert root_sm.__parent__ is root_folder
    assert root_sm.__bases__ == (component.getGlobalSiteManager(),)
    conn.add(root_sm)  # Ensure we have a connection so we can become KeyRefs
    assert root_sm._p_jar is conn

    root_folder.setSiteManager(root_sm)
    assert ISite.providedBy(root_folder)

    server_folder = main_factory()
    if not IMainApplicationFolder.providedBy(server_folder):
        interface.alsoProvides(server_folder, IMainApplicationFolder)
    conn.add(server_folder)
    root_folder[main_name] = server_folder
    assert server_folder.__parent__ is root_folder
    assert server_folder.__name__ == main_name
    assert root_folder[main_name] is server_folder

    lsm = BTreeLocalSiteManager(server_folder)
    conn.add(lsm)
    assert lsm.__parent__ is server_folder
    assert lsm.__bases__ == (root_sm,)

    server_folder.setSiteManager(lsm)
    assert ISite.providedBy(server_folder)

    with current_site(server_folder):
        assert component.getSiteManager() is lsm, "Component hooks must have been reset"

        # The name that many Zope components assume
        root[root_name] = root_folder
        if root_alias:
            root[root_alias] = server_folder

        root[server_folder.__name__] = server_folder
        if main_alias:
            root[main_alias] = server_folder

        if main_setup:
            main_setup(server_folder)

        lifecycleevent.added(root_folder)
        lifecycleevent.added(server_folder)

        install_sites_folder(server_folder)

    return root_folder, server_folder

def get_all_host_sites():
    """
    The order in which sites are accessed is top-down breadth-first,
    that is, the shallowest to the deepest nested sites. This allows
    you to assume that your parent sites have already been updated.

    :returns: A list of sites
    :rtype: list
    """

    sites = component.getUtility(IEtcNamespace, name='hostsites')
    sites = list(sites.values())

    # The easyiest way to go top-down is to again use the resolution order;
    # we just have to watch out for duplicates and non-persistent components
    site_to_ro = {site: ro.ro(site.getSiteManager()) for site in sites}

    # This should be a plain, directed acyclic tree (single root) that is now
    # linearized.
    # Transform from the site manager back into the site object itself
    site_to_site_ro = {}
    for site, managers in site_to_ro.items():
        site_to_site_ro[site] = [getattr(x, '__parent__', None) for x in managers]

    # Ok, now, go through the dictionary, walking from the top to the bottom,
    # one at a time, thus producing the correct order
    # (Because our datastructure looks like this:
    #   site1: [site1, ds, base, GSM]
    #   site2: [site2, site1, base, GSM]
    #   site3: [site3, ds, base, GSM])
    ordered = list()

    while site_to_site_ro:
        for site, managers in dict(site_to_site_ro).items():
            if not managers:
                site_to_site_ro.pop(site)
                continue

            base_site = managers.pop()
            if base_site in sites and base_site not in ordered:
                # Ie., it's a real one we haven't seen before
                ordered.append(base_site)
    return ordered

def run_job_in_all_host_sites(func):
    """
    While already operating inside of a transaction and the application
    environment, execute the callable given by ``func`` once for each
    persistent, registered host (see :func:`synchronize_host_policies`).
    The callable is run with that site current.

    This is typically used to make configuration changes/adjustments
    to utilities local within each site, while the appropriate event
    listeners for the site also fire.

    You are responsible for transaction management.

    :raises: Whatever the callable raises.
    :returns: A list of pairs `(site, result)` containing each site
        and the result of running the function in that site.
    :rtype: list
    """

    logger.debug("Asked to run job %s in ALL sites", func)

    results = list()
    ordered = get_all_host_sites()
    for site in ordered:
        results.append(run_job_in_host_site(site, func))
    return results

def get_host_site(site_name, safe=False):
    """
    Find the persistent site named *site_name* and return it.

    :keyword bool safe: If True, silently ignore any errors.
      **DO NOT** use this param. Deprecated and dangerous.
    """
    site = str(site_name)
    try:
        sites = component.getUtility(IEtcNamespace, name='hostsites')
        result = sites[site]
        return result
    except (ComponentLookupError, KeyError):
        if not safe:
            raise
    return None
get_site = get_host_site  # BWC

def run_job_in_host_site(site, func):
    """
    Helper function to run the *func* with the given site
    as the current site.

    :param site: Either the site object itself, or its unique name.
    """
    site = get_host_site(site) if isinstance(site, string_types) else site
    logger.debug('Running job %s in site %s', func, site.__name__)
    with current_site(site):
        result = func()
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""


.. $Id$
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

#disable: accessing protected members, too many methods
#pylint: disable=W0212,R0904


from hamcrest import assert_that
from hamcrest import is_


from nti.testing import base

import transaction

from zope import component
from ZODB.interfaces import IDatabase
from ZODB.Connection import Connection
import ZODB.DB
from ZODB.DemoStorage import DemoStorage
from zope.site import SiteManagerContainer

from ..runner import _connection_cm
from ..runner import _site_cm
from ..runner import run_job_in_site

from ..transient import TrivialSite

from ..interfaces import SiteNotInstalledError



class TestConnectionCM(base.AbstractTestBase):

    def test_connection_cm(self):
        db = ZODB.DB(DemoStorage(name='base'))
        component.provideUtility(db, IDatabase)

        with _connection_cm() as c:
            assert_that(c, is_(Connection))

class TestSiteCM(base.AbstractTestBase):

    def test_site_cm(self):
        class MockConn(object):
            def __init__(self):
                self._root = {}
            def root(self):
                return self._root

        c = MockConn()
        c._root['nti.dataserver'] = TrivialSite(component.getGlobalSiteManager())
        with _site_cm(c, ('abc',)) as sitemanc:
            assert_that(sitemanc, is_(c._root['nti.dataserver']))

    def test_site_cm_not_installed(self):
        class MockConn(object):
            def __init__(self):
                self._root = {}
            def root(self):
                return self._root

        c = MockConn()
        c._root['nti.dataserver'] = TrivialSite(None)
        try:
            with _site_cm(c, ('abc',)):
                self.fail("Should not get here")
        except SiteNotInstalledError:
            pass

class TestRunner(base.AbstractTestBase):

    def test_run(self):
        db = ZODB.DB(DemoStorage(name='base'))
        component.provideUtility(db, IDatabase)

        conn = db.open()
        smc = conn.root()['nti.dataserver'] = SiteManagerContainer()
        smc.setSiteManager(component.getGlobalSiteManager())

        transaction.commit()
        conn.close()

        def func():
            "A docstring"
            assert_that(component.getSiteManager(), is_(component.getGlobalSiteManager()))

        run_job_in_site(func, site_names=('abc',))

        run_job_in_site(func, job_name="Test")

        run_job_in_site(lambda: None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
.. $Id$
"""

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

import warnings

def registerUtility(registry, *args, **kwargs):
    return registry.registerUtility(*args, **kwargs)

def unregisterUtility(registry, *args, **kwargs):
    for k in 'event', 'force':
        if k in kwargs:
            kwargs.pop(k, None)
            warnings.warn(
                "unregisterUtility does not take '%s'; it will be dropped soon" % k,
                FutureWarning,
                stacklevel=2)
    return registry.unregisterUtility(*args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904

from hamcrest import is_
from hamcrest import is_not
from hamcrest import raises
from hamcrest import calling
from hamcrest import contains
from hamcrest import assert_that
from hamcrest import has_property
from hamcrest import same_instance
from hamcrest import none
does_not = is_not

import unittest

from zope import interface

from zope.interface import ro
from zope.interface import implementedBy
from zope.interface import Interface

from zope.component.hooks import getSite
from zope.component.hooks import setSite
from zope.component.hooks import site as currentSite

from zope.location.interfaces import LocationError

from zope.component import globalSiteManager as BASE

from z3c.baseregistry.baseregistry import BaseComponents

from nti.site.interfaces import IHostPolicyFolder
from nti.site.subscribers import threadSiteSubscriber
from nti.site.transient import HostSiteManager as HSM
from ..transient import TrivialSite
from ..site import get_site_for_site_names
from ..site import find_site_components
from ..site import get_component_hierarchy
from ..site import get_component_hierarchy_names

from nti.site.tests import SharedConfiguringTestLayer

from nti.testing.matchers import validly_provides
from nti.testing.matchers import is_false
from nti.testing.base import AbstractTestBase

from persistent import Persistent

import fudge

class IMock(Interface):
    pass

@interface.implementer(IMock)
class MockSite(object):

    __name__ = None
    __parent__ = None

    def __init__(self, site_man=None):
        self.site_man = site_man

    def getSiteManager(self):
        return self.site_man

class IFoo(Interface):
    pass

@interface.implementer(IFoo)
class RootFoo(object):
    pass

class TestSiteSubscriber(unittest.TestCase):

    layer = SharedConfiguringTestLayer

    def testProxyHostComps(self):
        pers_comps = BaseComponents(BASE, 'persistent', (BASE,))
        host_comps = BaseComponents(BASE, 'example.com', (BASE,))
        host_sm = HSM('example.com', 'siteman', host_comps, pers_comps)
        host_site = MockSite(host_sm)
        host_site.__name__ = host_sm.__name__
        setSite(host_site)

        new_comps = BaseComponents(BASE, 'sub_site', (pers_comps,))
        new_site = MockSite(new_comps)
        new_site.__name__ = new_comps.__name__
        interface.alsoProvides(new_site, IFoo)

        threadSiteSubscriber(new_site, None)

        cur_site = getSite()
        # It should implement the static and dynamic
        # ifaces
        assert_that(cur_site, validly_provides(IFoo))
        assert_that(cur_site, validly_provides(IMock))

        # It should have the marker property
        assert_that(cur_site.getSiteManager(),
                    has_property('host_components',
                                   host_comps))

        assert_that(ro.ro(cur_site.getSiteManager()),
                    contains(
                         # The first entry is synthesized
                         has_property('__name__', new_comps.__name__),
                         pers_comps,
                         # The host comps appear after all the bases
                         # in the ro of the new site
                         host_comps,
                         BASE))

    def testTraverseFailsIntoSiblingSiteExceptHostPolicyFolders(self):
        new_comps = BaseComponents(BASE, 'sub_site', ())
        new_site = MockSite(new_comps)
        new_site.__name__ = new_comps.__name__

        with currentSite(None):
            threadSiteSubscriber(new_site, None)
            # If we walk into a site...
            # ...and then try to walk into a sibling site with no apparent relationship...
            new_comps2 = BaseComponents(BASE, 'sub_site', (new_comps,))
            new_site2 = MockSite(new_comps2)
            new_site2.__name__ = new_comps2.__name__

            # ... we fail...
            assert_that(calling(threadSiteSubscriber).with_args(new_site2, None),
                        raises(LocationError))

            # ...unless they are both HostPolicyFolders...
            interface.alsoProvides(new_site, IHostPolicyFolder)
            interface.alsoProvides(new_site2, IHostPolicyFolder)
            threadSiteSubscriber(new_site2, None)

            # ... which does not change the site
            assert_that(getSite(), is_(same_instance(new_site)))

class TestGetSiteForSiteNames(AbstractTestBase):

    @fudge.patch('nti.site.site.find_site_components')
    def test_no_persistent_site(self, fake_find):
        pers_comps = BaseComponents(BASE, 'persistent', (BASE,))
        host_comps = BaseComponents(BASE, 'example.com', (BASE,))
        host_sm = HSM('example.com', 'siteman', host_comps, pers_comps)
        class PersistentTrivialSite(Persistent, TrivialSite):
            pass
        trivial_site = PersistentTrivialSite(host_sm)
        fake_find.is_callable().returns(pers_comps)


        x = get_site_for_site_names(('',), trivial_site)

        assert_that(x, is_not(Persistent))
        assert_that(x, is_(TrivialSite))
        assert_that(x.__name__, is_(pers_comps.__name__))

    def test_find_comps_empty(self):
        assert_that(find_site_components(('',)),
                    is_(none()))


class TestGetComponentHierarchy(AbstractTestBase):

    @fudge.patch('nti.site.site.find_site_components')
    def test_get_component_hierarchy(self, fake_find):
        # This is a lousy test, just testing what we already see done, not
        # correct behaviour
        pers_comps = BaseComponents(BASE, 'persistent', (BASE,))

        class MockSite(object):
            __parent__ = None
            __name__ = None

            def __init__(self):
                self.__parent__ = {}

        site_comps_1 = BaseComponents(pers_comps, '1', (pers_comps,))
        site_comps_2 = BaseComponents(site_comps_1, '2', (site_comps_1,))

        site = MockSite()
        site.__parent__[site_comps_1.__name__] = site_comps_1
        site.__parent__[site_comps_2.__name__] = site_comps_2

        fake_find.is_callable().returns(site_comps_2)

        x = list(get_component_hierarchy(site))
        assert_that(x, is_([site_comps_2, site_comps_1]))

        x = list(get_component_hierarchy_names(site))
        assert_that(x, is_(['2', '1']))

        x = list(get_component_hierarchy_names(site, reverse=True))
        assert_that(x, is_(['1', '2']))

from ..site import BTreeLocalSiteManager as BLSM
from ..site import _LocalAdapterRegistry
from ..site import BTreeLocalAdapterRegistry
from ..site import WrongRegistrationTypeError

from ZODB import DB
from ZODB.DemoStorage import DemoStorage
import transaction

import pickle
try:
    import zodbpickle.fastpickle as zpickle
except ImportError:
    import zodbpickle.pickle as zpickle # pypy
import BTrees.OOBTree

class TestBTreeSiteMan(AbstractTestBase):


    def test_pickle_setstate_swap_class(self):
        base_comps = BLSM(None)
        # replace with "broken"
        base_comps.adapters = _LocalAdapterRegistry()
        base_comps.utilities = _LocalAdapterRegistry()

        sub_comps = BLSM(None)
        sub_comps.__bases__ = (base_comps,)

        assert_that(sub_comps.adapters.__bases__, is_((base_comps.adapters,)))
        assert_that(sub_comps.utilities.__bases__, is_((base_comps.utilities,)))

        for p in pickle, zpickle:
            new_base, new_sub = p.loads(p.dumps([base_comps, sub_comps]))

            # Still in place
            assert_that(new_sub.adapters.__bases__, is_((new_base.adapters,)))
            assert_that(new_sub.utilities.__bases__, is_((new_base.utilities,)))

            # And they changed type
            assert_that(new_sub.adapters.__bases__[0], is_(BTreeLocalAdapterRegistry))
            assert_that(new_sub.utilities.__bases__[0], is_(BTreeLocalAdapterRegistry))


    def _store_base_subs_in_zodb(self, storage):
        db = DB(storage)
        conn = db.open()

        base_comps = BLSM(None)
        base_comps.btree_threshold = 0
        base_comps.__name__ = u'base'
        # replace with "broken"
        base_comps.adapters = _LocalAdapterRegistry()
        base_comps.utilities = _LocalAdapterRegistry()

        sub_comps = BLSM(None)
        sub_comps.__name__ = u'sub'
        sub_comps.__bases__ = (base_comps,)

        assert_that(sub_comps.adapters.__bases__, is_((base_comps.adapters,)))
        assert_that(sub_comps.utilities.__bases__, is_((base_comps.utilities,)))
        assert_that(sub_comps.utilities.__bases__[0], is_(_LocalAdapterRegistry))

        conn.root()['base'] = base_comps
        conn.root()['sub'] = sub_comps


        transaction.commit()
        conn.close()
        db.close()

    def test_pickle_setstate_swap_class_zodb(self):
        storage = DemoStorage()

        self._store_base_subs_in_zodb(storage)

        db = DB(storage)
        conn = db.open()
        new_base = conn.root()['base']
        new_base._p_activate()
        new_sub = conn.root()['sub']

        # Still in place
        assert_that(new_sub.adapters.__bases__, is_((new_base.adapters,)))
        assert_that(new_sub.utilities.__bases__, is_((new_base.utilities,)))

        # And they changed type
        assert_that(new_sub.adapters.__bases__[0], is_(BTreeLocalAdapterRegistry))
        assert_that(new_sub.utilities.__bases__[0], is_(BTreeLocalAdapterRegistry))

        # But p_changed wasn't set.
        assert_that(new_sub.adapters.__bases__[0],
                    has_property('_p_changed', is_false()))

        transaction.commit()
        conn.close()
        db.close()

        db = DB(storage)
        conn = db.open()
        new_sub = conn.root()['sub']
        # Now, we didn't rewrite the reference so the type stayed the same
        assert_that(type(new_sub.adapters.__bases__[0]), _LocalAdapterRegistry)
        assert_that(new_sub.adapters.__bases__[0],
                    has_property('_p_changed', is_false()))

        # Loading the base changes the types
        new_base = conn.root()['base']
        assert_that(new_base.adapters, is_(BTreeLocalAdapterRegistry))
        assert_that(new_base.adapters,
                    same_instance(new_sub.adapters.__bases__[0]))
        assert_that(new_sub.adapters.__bases__[0], is_(BTreeLocalAdapterRegistry))

        assert_that(new_sub.adapters.__bases__[0],
                    has_property('_p_changed', is_false()))

    def test_pickle_zodb_lookup_adapter(self):
        # Now, we can register a couple adapters in the base, save everything,
        # and look it up in the sub (when the classes don't match)
        storage = DemoStorage()
        self._store_base_subs_in_zodb(storage)

        db = DB(storage)
        conn = db.open()
        new_base = conn.root()['base']
        new_base._p_activate()
        new_sub = conn.root()['sub']


        new_base.adapters.btree_provided_threshold = 0
        new_base.adapters.btree_map_threshold = 1
        # Note: this causes btree-ing the map to fail. The implementedBy callable has default comparison
        # and can't be stored in a btree
        try:
            new_base.registerAdapter(_foo_factory,
                                     required=(object,),
                                     provided=IFoo)
            self.fail("Should raise")
        except WrongRegistrationTypeError:
            pass

        new_base.registerAdapter(_foo_factory2,
                                 required=(IFoo,),
                                 provided=IMock)

        assert_that(new_base._adapter_registrations, is_(BTrees.OOBTree.OOBTree))
        assert_that(new_base._adapter_registrations.keys(),
                    contains(
                        ((IFoo,), IMock, u''),
                    ))
        assert_that(new_base.adapters._provided, is_(BTrees.family64.OI.BTree))
        assert_that(new_base.adapters._adapters[0], is_({}))

        assert_that(new_base.adapters._adapters[1][IFoo], is_(dict))


        new_base.registerAdapter(_foo_factory2,
                                 required=(IFoo,),
                                 provided=IFoo)

        assert_that(new_base.adapters._adapters[1][IFoo], is_(BTrees.family64.OO.BTree))

        transaction.commit()
        conn.close()
        db.close()

        db = DB(storage)
        conn = db.open()
        new_sub = conn.root()['sub']

        x = new_sub.queryAdapter(RootFoo(), IMock)
        assert_that(x, is_(2))

    def test_pickle_zodb_lookup_utility(self):
        # Now, we can register a couple utilities in the base, save everything,
        # and look it up in the sub (when the classes don't match)
        storage = DemoStorage()
        self._store_base_subs_in_zodb(storage)

        db = DB(storage)
        conn = db.open()
        new_base = conn.root()['base']
        new_base._p_activate()
        new_sub = conn.root()['sub']


        new_base.utilities.btree_provided_threshold = 0
        new_base.utilities.btree_map_threshold = 0

        new_base.registerUtility(MockSite(),
                                 provided=IFoo)
        provided1 = new_base.adapters._provided
        try:
            new_base.registerUtility(MockSite(),
                                     provided=implementedBy(object),
                                     name=u'foo')
            self.fail("Should raise TypeError")
        except WrongRegistrationTypeError:
            # Once we've converted, we can't register implementedBy
            # again.
            pass

        new_base.registerUtility(MockSite(),
                                 provided=IMock,
                                 name=u'foo')

        provided2 = new_base.adapters._provided
        # Make sure that it only converted once
        assert_that(provided1, is_(same_instance(provided2)))
        assert_that(new_base._utility_registrations, is_(BTrees.OOBTree.OOBTree))

        assert_that(new_base._utility_registrations.keys(),
                    contains(
                        ((IFoo, u'')),
                        (IMock, u'foo'),
                    ))
        assert_that(new_base.utilities._provided, is_(BTrees.family64.OI.BTree))
        assert_that(new_base.utilities._adapters[0], is_(BTrees.family64.OO.BTree))

        assert_that(new_base.utilities._adapters[0][IFoo], is_(BTrees.family64.OO.BTree))


        transaction.commit()
        conn.close()
        db.close()

        db = DB(storage)
        conn = db.open()
        new_sub = conn.root()['sub']

        x = new_sub.queryUtility(IFoo)
        assert_that(x, is_(MockSite))

        x = new_sub.queryUtility(IMock, u'foo')
        assert_that(x, is_(MockSite))


    def test_convert_with_utility_registered_on_class(self):
        comps = BLSM(None)

        comps.utilities.btree_provided_threshold = 0
        comps.utilities.btree_map_threshold = 0

        assert_that(calling(comps.registerUtility).with_args(MockSite(),
                              provided=implementedBy(object),
                              name=u'foo'),
                    raises(WrongRegistrationTypeError))

    def test_convert_with_utility_no_provided(self):
        comps = BLSM(None)

        comps.utilities.btree_provided_threshold = 0
        comps.utilities.btree_map_threshold = 0

        class AUtility(object):
            # Doesn't implement any interfaces
            pass

        # You can't easily register them this way anyway
        assert_that(calling(comps.registerUtility).with_args(AUtility()),
                    raises(TypeError, "The utility doesn't provide a single interface"))

    def test_convert_with_utility_dynamic_provided(self):
        comps = BLSM(None)

        comps.btree_threshold = 0
        comps.utilities.btree_provided_threshold = 0
        comps.utilities.btree_map_threshold = 0

        class AUtility(object):
            # Doesn't implement any interfaces
            pass

        autility = AUtility()
        interface.alsoProvides(autility, IFoo)
        comps.registerUtility(autility)
        assert_that(comps._utility_registrations, is_(BTrees.OOBTree.OOBTree))
        assert_that(comps._utility_registrations.keys(),
                    contains(
                        ((IFoo, u'')),
                    ))
        assert_that(comps.utilities._provided, is_(BTrees.family64.OI.BTree))
        assert_that(comps.utilities._adapters[0], is_(BTrees.family64.OO.BTree))

        assert_that(comps.utilities._adapters[0][IFoo], is_(BTrees.family64.OO.BTree))

    def test_convert_with_adapter_registered_on_class(self):
        comps = BLSM(None)

        comps.btree_threshold = 0
        comps.adapters.btree_provided_threshold = 0
        comps.utilities.btree_map_threshold = 0

        assert_that(calling(comps.registerAdapter).with_args(
                           _foo_factory,
                           required=(object, str),
                           provided=IFoo),
                    raises(WrongRegistrationTypeError))


def _foo_factory(o):
    return 1
def _foo_factory2(o):
    return 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Implementations of sites and helpers for working with sites.

.. $Id$
"""
# NOTE: unicode_literals is NOT imported!!
from __future__ import print_function, absolute_import, division

__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

import sys

from zope.interface.interface import InterfaceClass
from zope import component

from zope.component.hooks import getSite

from zope.component.interfaces import IComponents

from persistent import Persistent

from nti.site.transient import TrivialSite
from nti.site.transient import HostSiteManager

_PYPY = hasattr(sys, 'pypy_version_info')
_DEFAULT_COMPARISON = "Can't use default __cmp__" if _PYPY else "Object has default comparison"

def find_site_components(site_names):
    """
    Return an (global, registered) :class:`.IComponents` implementation named
    for the first virtual site found in the sequence of *site_names*.
    If no such components can be found, returns none.
    """
    for site_name in site_names:
        if not site_name:  # Empty/default. We want the global. This should only ever be at the end
            return None
        components = component.queryUtility(IComponents, name=site_name)
        if components is not None:
            return components
_find_site_components = find_site_components  # BWC

def get_site_for_site_names(site_names, site=None):
    """
    Return an :class:`.ISite` implementation named for the first virtual site
    found in the sequence of site_names.

    If there is a registered persistent site having the same name as the
    registered global components found for *site_names*, then that site will be used.
    Otherwise, if there is only a registered global components, a non-persistent site
    that incorporates those components in the lookup order while still incorporating the
    current (or provided) site will be returned.

    If no such site or components can be found, returns the fallback
    site (the current or provided *site*).

    :param site_names: Sequence of strings giving the virtual host names
        to use.
    :keyword site: If given, this will be the fallback site (and site manager). If
        not given, then the currently installed site will be used.
    """

    if site is None:
        site = getSite()

    # assert site.getSiteManager().__bases__ == (component.getGlobalSiteManager(),)
    # Can we find a named site to use?
    site_components = find_site_components(site_names) if site_names else None  # micro-opt to not call if no names
    if site_components:
        # Yes we can.
        site_name = site_components.__name__
        # Do we have a persistent site installed in the database? If yes,
        # we want to use that.
        try:
            pers_site = site[u'++etc++hostsites'][site_name]
            site = pers_site
        except (KeyError, TypeError):
            # No, nothing persistent, dummy one up.
            # Note that this code path is deprecated now and not
            # expected to be hit.

            # The site components are only a
            # partial configuration and are not persistent, so we need
            # to use two bases to make it work (order matters) (for
            # example, the main site is almost always the
            # 'nti.dataserver' site, where the persistent intid
            # utilities live; the named sites do not have those and
            # cannot have the persistent nti.dataserver as their real
            # base, so the two must be mixed). They are also not
            # traversable.

            # Host comps used to be simple, but now they may be hierarchacl
            # assert site_components.__bases__ == (component.getGlobalSiteManager(),)
            # gsm = site_components.__bases__[0]
            # assert site_components.adapters.__bases__ == (gsm.adapters,)

            # But the current site, when given, must always be the main
            # dataserver site
            assert isinstance(site, Persistent)
            assert isinstance(site.getSiteManager(), Persistent)

            main_site = site
            site_manager = HostSiteManager(main_site.__parent__,
                                            main_site.__name__,
                                            site_components,
                                            main_site.getSiteManager())
            site = TrivialSite(site_manager)
            site.__parent__ = main_site
            site.__name__ = site_name

    return site

def get_component_hierarchy(site=None):
    site = getSite() if site is None else site
    # XXX: This is tightly coupled. Note that we assume that the parent
    # site is a container for the persistent sites.
    # There should never be a good reason to need to know this.
    hostsites = site.__parent__
    site_names = (site.__name__,)
    # XXX: Why is this not the same thing as site.getSiteManager()?
    components = find_site_components(site_names)
    while components is not None:
        try:
            name = components.__name__
            if name in hostsites:
                yield components
                components = components.__parent__
            else:
                break
        except AttributeError:  # pragma: no cover
            break

def get_component_hierarchy_names(site=None, reverse=False):
    # XXX This is tightly coupled and there should almost never
    # be a good reason to know this.
    result = [x.__name__ for x in get_component_hierarchy(site)]
    if reverse:
        result.reverse()
    return result

from zope.component.persistentregistry import PersistentComponents
from zope.site.site import LocalSiteManager
from zope.site.site import _LocalAdapterRegistry
from BTrees import family64

class WrongRegistrationTypeError(TypeError):
    """
    Raised if an adapter registration is of the wrong type.
    """


class _PermissiveOOBTree(family64.OO.BTree):

    def get(self, key, default=None):
        "Doesn't raise an error for getting something with default comparison."
        try:
            return super(_PermissiveOOBTree, self).get(key, default)
        except TypeError as e:
            if e.args[0] == _DEFAULT_COMPARISON:
                return default
            raise # pragma: no cover


class BTreeLocalAdapterRegistry(_LocalAdapterRegistry):
    """
    A persistent adapter registry that can switch its internal
    data structures to be more persistent friendly when they get large.

    .. caution:: This registry doesn't support registrations on bare
       classes. This is because the Implements and Provides objects
       returned on bare classes do not support comparison or equality
       and hence cannot be used in BTrees. (They only hash and compare
       equal to *themselves*; within the same process this works out
       because of aggressive caching on class objects.) Registering a utility
       to provide a bare class is quite hard to do, in any case. Registering
       adapters to require bare classes is easier but generally not a best practice.
    """
    # Inherit from _LocalAdapterRegistry for maximum compatibility...we are
    # going to swizzle out classes. Also, it makes sure we are ILocation.

    # Interestingly, we are totally fine to switch out the type from dict
    # to BTree. Much of the actual lookup code is implemented in C, but it calls
    # into Python for _uncached_lookup, which stays in pure python.

    #: The family for the provided map. Defaults to 64-bit maps. I.e., long.
    btree_family = family64

    #: The type of BTree to be used for adapter registrations. This generally shouldn't
    #: be changed. In an emergency, it can be set to :class:`dict` to avoid doing any
    #: migrations.
    btree_oo_type = _PermissiveOOBTree

    #: The size at which the total number of registered adapters will
    #: be switched to a BTree.
    btree_provided_threshold = 5000

    #: The size at which individual keys in the lookup decision maps
    #: will be switched to BTrees. The map threshold is lower than the provided threshold
    #: because it is there are many keys in the map so the overall
    #: effect is amplified.
    btree_map_threshold = 2000

    # REMEMBER: Always check the type *before* checking the length.
    # Getting the length of a bare BTree is expensive and loads all the
    # buckets into memory.

    def _check_and_btree_maps(self, byorder):
        btree_type = self.btree_oo_type
        for i in range(len(byorder)):
            mapping = byorder[i]
            if not isinstance(mapping, btree_type) and len(mapping) > self.btree_map_threshold:
                mapping = btree_type(mapping)
                byorder[i] = mapping
                # self._adapters and self._subscribers are both simply
                # of type `list` (not persistent list) so when we make changes
                # to them, we need to set self._p_changed
                self._p_changed = True

            # This is the first level of the decision tree, and thus
            # the least discriminatory. If i is 0, then this is only
            # things that are specifically providing a single interface
            # (Which is the most common in some usages). These maps are thus
            # liable to get to be the biggest. Note that we only replace at this
            # level.
            replacement_vals = {}
            for k, v in mapping.items():
                if not isinstance(v, btree_type) and len(v) > self.btree_map_threshold:
                    replacement_vals[k] = btree_type(v)

            if replacement_vals:
                mapping.update(replacement_vals)
                if not isinstance(mapping, btree_type):
                    # This may or may not be a btree, depending on its own size,
                    # so we may need to mark ourself as changed.
                    self._p_changed = True

    def register(self, required, provided, name, value):
        """
        Raises :exc:`WrongRegistrationTypeError` if *required* or
        *provided* contains a bare class.
        """
        # Pre-check that there are no bad registrations.
        for r in required:
            if r is not None and not isinstance(r, InterfaceClass):
                raise WrongRegistrationTypeError("Unsupported declaration type in required %s" % r)
        if not isinstance(provided, InterfaceClass):
            raise WrongRegistrationTypeError("Unsupported declaration type in provided %s" % provided)
        return super(BTreeLocalAdapterRegistry, self).register(required, provided, name, value)

    def changed(self, originally_changed):
        # If we changed, check and migrate
        if originally_changed is self:
            if (not isinstance(self._provided, self.btree_family.OI.BTree)
                and len(self._provided) > self.btree_provided_threshold):
                self._provided = self.btree_family.OI.BTree(self._provided)
                self._p_changed = True
            for byorder in self._adapters, self._subscribers:
                self._check_and_btree_maps(byorder)
        super(BTreeLocalAdapterRegistry, self).changed(originally_changed)

class BTreePersistentComponents(PersistentComponents):
    """
    Persistent components that will be friendly to ZODB when they get large.

    Note that despite the name, this class is not Persistent, only its
    internal components are.

    .. caution:: This registry doesn't support bare class registrations.
       See :class:`BTreeLocalAdapterRegistry` for details.
    """

    btree_family = family64

    #: The size at which we will switch from maps to BTrees for registered adapters
    #: and registered utilities (individually).
    btree_threshold = 5000

    def _init_registries(self):
        # NOTE: We cannot simply replace these two attributes at runtime
        # or even in a migration (for example, to upgrade from one type to another type)
        # and expect it to work. If we are the base of some other Components
        # or SiteManager, then these attributes have been copied into the __bases__
        # of *its* adapters and utilities. If we swap out our ivar, then the bases
        # will be out of sync and lookup will be broken. (BTreeLocalSiteManager
        # supposedly keeps track of its subs and so it *could* swap out all of them too.)
        self.adapters = BTreeLocalAdapterRegistry()
        self.utilities = BTreeLocalAdapterRegistry()
        self.adapters.__parent__ = self.utilities.__parent__ = self
        self.adapters.__name__ = u'adapters'
        self.utilities.__name__ = u'utilities'

    def _check_and_btree_map(self, mapping_name):
        btree_type = self.btree_family.OO.BTree
        mapping = getattr(self, mapping_name)
        if not isinstance(mapping, btree_type) and len(mapping) > self.btree_threshold:
            mapping = btree_type(mapping)
            setattr(self, mapping_name, mapping)
            # NOTE: This class is *NOT* Persistent, but its subclass BTreeLocalSiteManager
            # *is*. That's why __setstate__ is there and not here...it doesn't make much sense here.

    def registerUtility(self, component=None, provided=None, name=u'', info=u'',
                        event=True, factory=None):
        result = None
        try:
            result = super(BTreePersistentComponents, self).registerUtility(
                component, provided, name, info, event, factory)
        except WrongRegistrationTypeError:
            # Back out our partial state changes
            self.unregisterUtility(component, provided, name, factory)
            raise
        self._check_and_btree_map('_utility_registrations')

        return result

    def registerAdapter(self, *args, **kwargs):
        result = None
        try:
            result = super(BTreePersistentComponents, self).registerAdapter(*args, **kwargs)
        except WrongRegistrationTypeError:
            kwargs.pop('event', None)
            self.unregisterAdapter(*args, **kwargs)
            raise
        self._check_and_btree_map('_adapter_registrations')
        return result

class BTreeLocalSiteManager(BTreePersistentComponents, LocalSiteManager):
    """
    Persistent local site manager that will be friendly to ZODB when they
    get large.

    .. caution:: This registry doesn't support bare class registrations.
       See :class:`BTreeLocalAdapterRegistry` for details.
    """

    def __setstate__(self, state):
        super(BTreeLocalSiteManager, self).__setstate__(state)
        # Graceful migration from older versions of this class.
        # See note in _init_registries for why we can't simply swap these to new
        # ivars. Instead, we adjust their __class__. Note that we'll have to keep doing this
        # forever or until we save a brand new copy of the object, because the class is stored
        # as part of the pickle. Adjusting the class works because we know that the layout
        # is exactly the same. Now, other objects could be awake and active and querying
        # this object under its old class through their own __bases__, but that's ok:
        # our behaviour modification only comes in at write time...which only happens
        # through methods we expose, so we'll get a chance to swizzle the object out.
        for reg in self.adapters, self.utilities:
            if (not isinstance(reg, BTreeLocalAdapterRegistry)
                and isinstance(reg, _LocalAdapterRegistry)):
                # Only do this for classes we know about.
                # Note: In Persistent 4.2.1, pure-python and C handle __class__ differently.
                # Pure-python doesn't set _p_changed, but C does.
                changed = reg._p_changed
                reg.__class__ = BTreeLocalAdapterRegistry
                if not changed:
                    reg._p_changed = False


# Legacy notes:
# Opening the connection registered it with the transaction manager as an ISynchronizer.
# Ultimately this results in newTransaction being called on the connection object
# at `transaction.begin` time, which in turn syncs the storage. However,
# when multi-databases are used, the other connections DO NOT get this called on them
# if they are implicitly loaded during the course of object traversal or even explicitly
# loaded by name turing an active transaction. This can lead to extra read conflict errors
# (particularly with RelStorage which explicitly polls for invalidations at sync time).
# (Once a multi-db connection has been used, then the next time it would be sync'd. A multi-db
# connection is associated with the same connection to another database for its lifetime, and
# when open()'d will sync all other such connections. Corrollary: ALWAYS go through
# a connection object to get access to multi databases; never go through the database object itself.)

# As a workaround, we iterate across all the databases and sync them manually; this increases the
# cost of handling transactions for things that do not use the other connections, but ensures
# we stay nicely in sync.

# JAM: 2012-09-03: With the database resharding, evaluating the need for this.
# Disabling it.
# for db_name, db in conn.db().databases.items():
#   __traceback_info__ = i, db_name, db, func
#   if db is None: # For compatibility with databases we no longer use
#       continue
#   c2 = conn.get_connection(db_name)
#   if c2 is conn:
#       continue
#   c2.newTransaction()

# Now fire 'newTransaction' to the ISynchronizers, including the root connection
# This may result in some redundant fires to sub-connections.



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Site interfaces.

.. $Id$
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

from zope import interface

from zope.site.interfaces import IFolder
from zope.site.interfaces import ILocalSiteManager

from nti.schema.field import Number

class InappropriateSiteError(LookupError):
    """
    Raised if the installed site is not suitable to work with.

    Currently no code in this package raises this exception.
    """

class SiteNotInstalledError(AssertionError):
    """
    Raised when setting and getting a site do not work.

    This is most often caused by not installing the zope.component hooks.
    """

class IMainApplicationFolder(IFolder):
    """
    The folder representing the application. The set of persistent
    components will be installed beneath this folder, and this
    folder will be an Site (with a site manager).

    This may be the same thing as the root folder. As an implementation
    note, though, this is typically beneath the root folder and named
    something application specific (e.g., "dataserver2").
    """
class IHostPolicyFolder(IFolder):
    """
    A folder that should always have a site manager, and thus is a
    site, representing a policy for the host name. Persistent
    configuration related to that host should reside in this folder.
    """

class IHostPolicySiteManager(ILocalSiteManager):
    """
    A persistent local site manager that is tied to a site name. It should always
    have two bases, a non-persistent global IComponents configured through
    :mod:`z3c.baseregistry` and the persistent main dataserver site manager,
    in that order. This should be the site manager for an :class:`IHostPolicyFolder`
    """

class IHostSitesFolder(IFolder):
    """
    A container for the sites, each of which should be an
    :class:`IHostPolicyFolder`
    """

    lastSynchronized = Number(title=u"The timestamp at which this object was last synchronized .",
                              default=0.0)
    lastSynchronized.setTaggedValue('_ext_excluded_out', True)

class ITransactionSiteNames(interface.Interface):
    """
    An interface for a utility that return a list of the possible site names
    needed whenever transaction is run.
    """

    def __call__(*args, **kwargs):
        "Return the possible site names"

class ISiteTransactionRunner(interface.Interface):
    """
    Something that runs code within a transaction, properly setting up
    the ZODB, persistent site, and its environment.
    """

    def __call__(func, retries=0, sleep=None, site_names=(), side_effect_free=False,
                 root_folder_name='nti.dataserver'):
        """
        Runs the function given in `func` in a transaction and application local
        site manager (defaulting to the current site manager).

        :param function func: A function of zero parameters to run. If
            it has a docstring, that will be used as the transactions
            note. A transaction will be begun before this function
            executes, and committed after the function completes. This
            function may be rerun if retries are requested, so it
            should be prepared for that.

        :keyword int retries: The number of times to retry the
            transaction and execution of `func` if
            :class:`transaction.interfaces.TransientError` is raised
            when committing. Defaults to zero (so the job runs once).
            If you specify None, an implementation-specific
            number of retries will be used.

        :keyword float sleep: If not none, then the greenlet running
            this function will sleep for this long between retry
            attempts.

        :keyword site_names: DEPRECATED. Sequence of strings giving the
            virtual host names to use. If you do not provide this argument,
            then the currently installed site will be maintained when
            the transaction is run. NOTE: The implementation of this
            may maintain either the site names or the actual site
            object.

        :keyword bool side_effect_free: If true (not the default), then
            the function is assummed to have no side effects that need
            to be committed; the transaction runner is free to abort/rollback
            or commit the transaction at its leisure.

        :keyword str root_folder_name: This names the folder that can be found in the
            root of the ZODB that will serve as the starting point to look for the
            persistent named site.

        :return: The value returned by the first successful invocation of `func`.
        """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Implementations of folders for site policies.

.. $Id$
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

from zope import interface

from zope.site.folder import Folder

from .site import BTreeLocalSiteManager

from ZODB.POSException import ConnectionStateError

from nti.site.interfaces import IHostSitesFolder
from nti.site.interfaces import IHostPolicyFolder
from nti.site.interfaces import IHostPolicySiteManager

@interface.implementer(IHostSitesFolder)
class HostSitesFolder(Folder):
    """
    Simple container implementation for named host sites.
    """
    lastSynchronized = 0

    def __repr__(self):
        try:
            return super(HostSitesFolder, self).__repr__()
        except ConnectionStateError:
            return object.__repr__(self)

@interface.implementer(IHostPolicyFolder)
class HostPolicyFolder(Folder):
    """
    Simple container implementation for the named host site.
    """

    def __str__(self): # pragma: no cover
        return 'HostPolicyFolder(%s)' % self.__name__

    def __repr__(self):
        return 'HostPolicyFolder(%s,%s)' % (self.__name__, id(self))

@interface.implementer(IHostPolicySiteManager)
class HostPolicySiteManager(BTreeLocalSiteManager):

    def __repr__(self):
        try:
            return super(HostPolicySiteManager, self).__repr__()
        except ConnectionStateError:
            return object.__repr__(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904

from hamcrest import is_
from hamcrest import any_of
from hamcrest import is_not
from hamcrest import raises
from hamcrest import calling
from hamcrest import has_key
from hamcrest import contains
from hamcrest import not_none
from hamcrest import has_length
from hamcrest import assert_that
from hamcrest import has_property
from hamcrest import same_instance
does_not = is_not

import unittest

from zope import interface

from zope.interface import ro
from zope.interface import Interface

from zope import component

from zope.component.hooks import getSite
from zope.component.hooks import setSite
from zope.component.hooks import site as currentSite

from zope.location.interfaces import LocationError

from nti.site.interfaces import IHostPolicyFolder
from nti.site.subscribers import threadSiteSubscriber
from nti.site.transient import HostSiteManager as HSM

from nti.site.tests import SharedConfiguringTestLayer

from nti.testing.matchers import validly_provides

class IMock(Interface):
    pass

@interface.implementer(IMock)
class MockSite(object):

    __parent__ = None
    __name__ = None
    def __init__(self, site_man=None):
        self.site_man = site_man

    def getSiteManager(self):
        return self.site_man

from zope.component import globalSiteManager as BASE

from z3c.baseregistry.baseregistry import BaseComponents

class IFoo(Interface):
    pass

class TestSiteSubscriber(unittest.TestCase):

    layer = SharedConfiguringTestLayer

    def testProxyHostComps(self):
        pers_comps = BaseComponents(BASE, 'persistent', (BASE,))
        host_comps = BaseComponents(BASE, 'example.com', (BASE,))
        host_sm = HSM('example.com', 'siteman', host_comps, pers_comps)
        host_site = MockSite(host_sm)
        host_site.__name__ = host_sm.__name__
        setSite(host_site)

        new_comps = BaseComponents(BASE, 'sub_site', (pers_comps,))
        new_site = MockSite(new_comps)
        new_site.__name__ = new_comps.__name__
        interface.alsoProvides(new_site, IFoo)

        threadSiteSubscriber(new_site, None)

        cur_site = getSite()
        # It should implement the static and dynamic
        # ifaces
        assert_that(cur_site, validly_provides(IFoo))
        assert_that(cur_site, validly_provides(IMock))

        # It should have the marker property
        assert_that(cur_site.getSiteManager(),
                    has_property('host_components',
                                   host_comps))

        assert_that(ro.ro(cur_site.getSiteManager()),
                    contains(
                         # The first entry is synthesized
                         has_property('__name__', new_comps.__name__),
                         pers_comps,
                         # The host comps appear after all the bases
                         # in the ro of the new site
                         host_comps,
                         BASE))

    def testTraverseFailsIntoSiblingSiteExceptHostPolicyFolders(self):
        new_comps = BaseComponents(BASE, 'sub_site', ())
        new_site = MockSite(new_comps)
        new_site.__name__ = new_comps.__name__

        with currentSite(None):
            threadSiteSubscriber(new_site, None)
            # If we walk into a site...

            # ...and then try to walk into a sibling site with no apparent relationship...
            new_comps2 = BaseComponents(BASE, 'sub_site', (new_comps,))
            new_site2 = MockSite(new_comps2)
            new_site2.__name__ = new_comps2.__name__

            # ... we fail...
            assert_that(calling(threadSiteSubscriber).with_args(new_site2, None),
                        raises(LocationError))

            # ...unless they are both HostPolicyFolders...
            interface.alsoProvides(new_site, IHostPolicyFolder)
            interface.alsoProvides(new_site2, IHostPolicyFolder)
            repr(new_site) # coverage
            str(new_site)
            threadSiteSubscriber(new_site2, None)

            # ... which does not change the site
            assert_that(getSite(), is_(same_instance(new_site)))

# Match a hierarchy we have in nti.app.sites.demo:
# global
#  \
#   eval
#   |\
#   | eval-alpha
#   \
#   demo
#    \
#     demo-alpha

EVAL = BaseComponents(BASE,
                      name='eval.nextthoughttest.com',
                      bases=(BASE,))

EVALALPHA = BaseComponents(EVAL,
                           name='eval-alpha.nextthoughttest.com',
                           bases=(EVAL,))

DEMO = BaseComponents(EVAL,
                      name='demo.nextthoughttest.com',
                      bases=(EVAL,))

DEMOALPHA = BaseComponents(DEMO,
                           name='demo-alpha.nextthoughttest.com',
                           bases=(DEMO,))

_SITES = (EVAL, EVALALPHA, DEMO, DEMOALPHA)

from zope.component.interfaces import ISite
from zope.component.interfaces import IComponents

from zope.site.interfaces import INewLocalSite

from nti.site.interfaces import IHostPolicySiteManager

from nti.site.hostpolicy import synchronize_host_policies
from nti.site.hostpolicy import run_job_in_all_host_sites
from nti.site.hostpolicy import get_host_site

from nti.site.site import _find_site_components
from nti.site.site import get_site_for_site_names

from nti.site.tests import WithMockDS
from nti.site.tests import mock_db_trans

from nti.testing.matchers import verifiably_provides

class ITestSiteSync(interface.Interface):
    pass

@interface.implementer(ITestSiteSync)
class ASync(object):
    pass

@interface.implementer(ITestSiteSync)
class OtherSync(object):
    pass

class TestSiteSync(unittest.TestCase):

    layer = SharedConfiguringTestLayer

    _events = ()

    def setUp(self):
        super(TestSiteSync, self).setUp()
        for site in _SITES:
            # See explanation in nti.appserver.policies.sites; in short,
            # the teardown process can disconnect the resolution order of
            # these objects, and since they don't descend from the bases declared
            # in that module, they fail to get reset.
            site.__init__(site.__parent__, name=site.__name__, bases=site.__bases__)
            BASE.registerUtility(site, name=site.__name__, provided=IComponents)
        self._events = []
        # NOTE: We can't use an instance method; under
        # zope.testrunner, by the time tearDown is called, it's not
        # equal to the value it has during setUp, so we can't
        # unregister it!
        self._event_handler = lambda *args: self._events.append(args)
        BASE.registerHandler(self._event_handler, required=(IHostPolicySiteManager, INewLocalSite))
        DEMO.registerUtility(ASync(), provided=ITestSiteSync)

    def tearDown(self):
        for site in _SITES:
            BASE.unregisterUtility(site, name=site.__name__, provided=IComponents)
        BASE.unregisterHandler(self._event_handler, required=(IHostPolicySiteManager, INewLocalSite))
        super(TestSiteSync, self).tearDown()

    def test_simple_ro(self):
        # Check that resolution order is what we think. See
        # site.py
        # This simulates the layout in the database and global
        # site manager.
        class GSM(object): pass
        # DB
        class Root(GSM): pass
        class DS(Root): pass
        # global sites
        class Base(GSM): pass
        class S1(Base): pass
        class S2(Base): pass
        # DB sites
        class PS1(S1, DS): pass
        class PS2(S2, PS1): pass

        assert_that(ro.ro(PS2),
                     is_([PS2, S2, PS1, S1, Base, DS, Root, GSM, object]))

    @WithMockDS
    def test_site_sync(self):

        for site in _SITES:
            assert_that(_find_site_components((site.__name__,)),
                         is_(not_none()))

        with mock_db_trans() as conn:
            for site in _SITES:
                assert_that(_find_site_components((site.__name__,)),
                             is_(not_none()))


            ds = conn.root()['nti.dataserver']
            assert ds is not None
            sites = ds['++etc++hostsites']
            for site in _SITES:
                assert_that(sites, does_not(has_key(site.__name__)))

            synchronize_host_policies()
            synchronize_host_policies()

            assert_that(self._events, has_length(len(_SITES)))
            # These were put in in order
            # assert_that( self._events[0][0].__parent__,
            #            has_property('__name__', EVAL.__name__))

            # XXX These two lines are cover only.
            get_host_site(DEMO.__name__)
            get_host_site('DNE', True)
            assert_that(calling(get_host_site).with_args('dne'),
                        raises(LookupError))

        with mock_db_trans() as conn:
            for site in _SITES:
                assert_that(_find_site_components((site.__name__,)),
                             is_(not_none()))

            ds = conn.root()['nti.dataserver']

            assert ds is not None
            sites = ds['++etc++hostsites']

            assert_that(sites, has_key(EVAL.__name__))
            assert_that(sites[EVAL.__name__], verifiably_provides(ISite))

            # If we ask the demoalpha persistent site for an ITestSyteSync,
            # it will find us, because it goes to the demo global site
            assert_that(sites[DEMOALPHA.__name__].getSiteManager().queryUtility(ITestSiteSync),
                         is_(ASync))

            # However, if we put something in the demo *persistent* site, it
            # will find that
            sites[DEMO.__name__].getSiteManager().registerUtility(OtherSync())
            assert_that(sites[DEMOALPHA.__name__].getSiteManager().queryUtility(ITestSiteSync),
                        is_(OtherSync))

            # Verify the resolution order too
            def _name(x):
                if x.__name__ == '++etc++site':
                    return 'P' + str(x.__parent__.__name__)
                return x.__name__
            assert_that([_name(x) for x in ro.ro(sites[DEMOALPHA.__name__].getSiteManager())],
                        is_([u'Pdemo-alpha.nextthoughttest.com',
                             u'demo-alpha.nextthoughttest.com',
                             u'Pdemo.nextthoughttest.com',
                             u'demo.nextthoughttest.com',
                             u'Peval.nextthoughttest.com',
                             u'eval.nextthoughttest.com',
                             u'Pdataserver2',
                             u'PNone',
                             'base']))

            # including if we ask to travers from top to bottom
            names = list()
            def func():
                names.append(_name(component.getSiteManager()))

            run_job_in_all_host_sites(func)
            # Note that PDemo and Peval-alpha are arbitrary, they both
            # descend from eval;
            # TODO: why aren't we maintaining alphabetical order?
            # we should be, but sometimes we don't
            assert_that(names, is_(any_of(
                [u'Peval.nextthoughttest.com',
                 u'Pdemo.nextthoughttest.com',
                 u'Peval-alpha.nextthoughttest.com',
                 u'Pdemo-alpha.nextthoughttest.com'],
                [u'Peval.nextthoughttest.com',
                 u'Peval-alpha.nextthoughttest.com',
                 u'Pdemo.nextthoughttest.com',
                 u'Pdemo-alpha.nextthoughttest.com'])))

            # And that it's what we get back if we ask for it
            assert_that(get_site_for_site_names((DEMOALPHA.__name__,)),
                         is_(same_instance(sites[DEMOALPHA.__name__])))

        # No new sites created
        assert_that(self._events, has_length(len(_SITES)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904

from hamcrest import is_
from hamcrest import is_not
from hamcrest import raises
from hamcrest import calling
from hamcrest import assert_that
from hamcrest import has_property
does_not = is_not

import unittest

import pickle

from zope.component import globalSiteManager as BASE

from z3c.baseregistry.baseregistry import BaseComponents

from nti.site.transient import HostSiteManager as HSM
from nti.site.transient import TrivialSite

class TestHSM(unittest.TestCase):

    def _makeOne(self):
        pers_comps = BaseComponents(BASE, 'persistent', (BASE,))
        host_comps = BaseComponents(BASE, 'example.com', (BASE,))
        host_sm = HSM('example.com', 'siteman', host_comps, pers_comps)

        return host_sm, pers_comps, host_comps

    def test_cover(self):
        host_sm, pers_comps, host_comps = self._makeOne()

        assert_that(host_sm, has_property('host_components', is_(host_comps)))
        assert_that(host_sm, has_property('persistent_components', is_(pers_comps)))

    def test_pickle(self):
        host_sm = self._makeOne()[0]
        assert_that(calling(pickle.dumps).with_args(host_sm),
                    raises(TypeError, "BasedSiteManager should not be pickled"))

    def test_not_folderish(self):
        host_sm = self._makeOne()[0]
        assert_that(calling(host_sm.__setitem__).with_args('key', object()),
                    raises(AttributeError))


class TestTrivialSite(unittest.TestCase):

    def _makeOne(self):
        pers_comps = BaseComponents(BASE, 'persistent', (BASE,))
        host_comps = BaseComponents(BASE, 'example.com', (BASE,))
        host_sm = HSM('example.com', 'siteman', host_comps, pers_comps)

        site = TrivialSite(host_sm)
        return site, host_sm

    def test_no_pickle(self):
        site, _ = self._makeOne()
        assert_that(calling(pickle.dumps).with_args(site),
                    raises(TypeError, "TrivialSite should not be pickled"))

    def test_site_manager(self):
        site, sm = self._makeOne()
        assert_that(site.getSiteManager(), is_(sm))



import codecs
from setuptools import setup, find_packages



entry_points = {
    'console_scripts': [
    ],
}

TESTS_REQUIRE = [
    'fudge',
    'nose2[coverage_plugin]',
    'nti.testing',
    'pyhamcrest',
    'z3c.baseregistry',
    'zope.testrunner',
]

def _read(fname):
    with codecs.open(fname, encoding='utf-8') as f:
        return f.read()

setup(
    name='nti.site',
    version=_read('version.txt').strip(),
    author='Jason Madden',
    author_email='jason@nextthought.com',
    description="NTI Site",
    long_description=_read('README.rst'),
    license='Apache',
    keywords='Site management',
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    zip_safe=True,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    namespace_packages=['nti'],
    tests_require=TESTS_REQUIRE,
    install_requires=[
        'setuptools',
        'persistent',
        'six',
        'ZODB',
        'zope.component',
        'zope.container',
        'zope.interface',
        'zope.location',
        'zope.proxy',
        'zope.site',
        'zope.traversing',
        'nti.schema',
        'nti.transactions'
    ],
    extras_require={
        'test': TESTS_REQUIRE,
    },
    entry_points=entry_points,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Transient, in-memory, non-persistent site and site manager
implementations. These are used to get non-persistent
host-based global IComponents into the base resolution order.

.. $Id$
"""

# turn off warning for not calling superclass, calling indirect superclass and
# accessing protected methods. we're deliberately doing both
# pylint: disable=W0233,W0231,W0212

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

from zope import interface

from zope.component import interfaces as comp_interfaces
from zope.component.persistentregistry import PersistentComponents as _ZPersistentComponents

from zope.container.contained import Contained as _ZContained

from zope.site.site import LocalSiteManager as _ZLocalSiteManager

# TODO: All this site mucking may be expensive. It has significant possibilities
# for optimization (caching) using the fact that much of it is read only.

class BasedSiteManager(_ZLocalSiteManager):
    """
    A site manager that exists simply to have bases, but not to
    record itself as children of those bases (since that's unnecessary
    for our purposes and leads to ZODB conflicts).
    """

    # Note that the adapter registries in the base objects /will/ have
    # weak references to this object; it's very hard to stop this. These
    # will stick around until a gc is run. (For testing purposes,
    # it is important to GC or you can get weird errors like:
    # File "zope/interface/adapter.py", line 456, in changed
    #   super(AdapterLookupBase, self).changed(None)
    # File "ZODB/Connection.py", line 857, in setstate
    #   raise ConnectionStateError(msg)
    #  ConfigurationExecutionError: <class 'ZODB.POSException.ConnectionStateError'>:
    #   Shouldn't load state for 0x237f4ee301650a49 when the connection is closed
    # in:
    # File "zope/site/configure.zcml", line 13.4-14.71
    # <implements interface="zope.annotation.interfaces.IAttributeAnnotatable" />
    # Fortunately, Python's GC is precise and refcounting, so as long as we do not leak
    # refs to these, we're fine

    def _setBases(self, bases):
        # Bypass the direct superclass.
        _ZPersistentComponents._setBases(self, bases)

    def __init__(self, site, name, bases):
        # Bypass the direct superclass to avoid setting
        # bases multiple times and initing the BTree portion, which we won't use
        # NOTE: This means we are fairly tightly coupled
        _ZPersistentComponents.__init__(self)

        # Locate the site manager
        self.__parent__ = site
        self.__name__ = name
        self.__bases__ = bases

    def _newContainerData(self):  # pragma: no cover
        return None  # We won't be used as a folder

    def __reduce__(self):
        raise TypeError("BasedSiteManager should not be pickled")

    __getstate__ = __reduce__


class HostSiteManager(BasedSiteManager):
    """
    A site manager that is intended to be used with globally
    registered IComponents plus the application persistent components.
    """

    def __init__(self, site, name, host_components, persistent_components):
        self._host_components = host_components
        self._persistent_components = persistent_components
        BasedSiteManager.__init__(self,
                                  site,
                                  name,
                                  (host_components, persistent_components))

    @property
    def host_components(self):
        return self._host_components

    @property
    def persistent_components(self):
        return self._persistent_components

@interface.implementer(comp_interfaces.ISite)
class TrivialSite(_ZContained):
    """
    Trivial non-persistent implementation of :class:`.ISite`
    """
    def __init__(self, site_manager):
        self._sm = site_manager

    def getSiteManager(self):
        return self._sm

    def __reduce__(self):
        raise TypeError("TrivialSite should not be pickled")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, unicode_literals, absolute_import, division

__docformat__ = "restructuredtext en"

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904

logger = __import__('logging').getLogger(__name__)

import functools
import transaction


from zope import component

from zope.component.hooks import setHooks
from zope.component.hooks import site as currentSite



from ..site import BTreeLocalSiteManager
from ..hostpolicy import install_main_application_and_sites
from zope.site import SiteManagerContainer

import ZODB

from ZODB.DemoStorage import DemoStorage

from ..site import get_site_for_site_names

from nti.testing.layers import find_test
from nti.testing.layers import GCLayerMixin
from nti.testing.layers import ZopeComponentLayer
from nti.testing.layers import ConfiguringLayerMixin

from hamcrest import assert_that
from nti.testing.matchers import provides
from ..interfaces import IMainApplicationFolder

import zope.testing.cleanup

current_mock_db = None
current_transaction = None

root_name = u'nti.dataserver'

def install_main(conn):
    def setup(f):
        assert_that(f, provides(IMainApplicationFolder))
    install_main_application_and_sites(conn,
                                       root_alias=root_name,
                                       main_name=u'dataserver2',
                                       main_setup=setup)

def init_db(db, conn=None):
    conn = db.open() if conn is None else conn
    global current_transaction
    if current_transaction != conn:
        current_transaction = conn
    install_main(conn)
    return conn

class mock_db_trans(object):

    def __init__(self, db=None, site_name=None):
        self.db = db or current_mock_db
        self._site_cm = None
        self._site_name = site_name

    def _check(self, conn):
        root = conn.root()
        if root_name not in root:
            install_main(conn)

    def __enter__(self):
        transaction.begin()
        self.conn = conn = self.db.open()
        global current_transaction
        current_transaction = conn
        self._check(conn)

        sitemanc = conn.root()[root_name]
        if self._site_name:
            with currentSite(sitemanc):
                sitemanc = get_site_for_site_names((self._site_name,))

        self._site_cm = currentSite(sitemanc)
        self._site_cm.__enter__()
        assert component.getSiteManager() == sitemanc.getSiteManager()
        return conn

    def __exit__(self, t, v, tb):
        result = self._site_cm.__exit__(t, v, tb)
        global current_transaction
        body_raised = t is not None
        try:
            try:
                if not transaction.isDoomed():
                    transaction.commit()
                else:
                    transaction.abort()
            except Exception:
                transaction.abort()
                raise
            finally:
                current_transaction = None
                self.conn.close()
        except Exception:
            if not body_raised:
                raise
            logger.exception("Failed to cleanup trans, but body raised exception too")
        reset_db_caches(self.db)
        return result

def reset_db_caches(db=None):
    if db is not None:
        db.pool.map(lambda conn: conn.cacheMinimize())

def _mock_ds_wrapper_for(func, db, teardown=None):

    @functools.wraps(func)
    def f(*args):
        global current_mock_db
        current_mock_db = db
        init_db(db)

        sitemanc = SiteManagerContainer()
        sitemanc.setSiteManager(BTreeLocalSiteManager(None))

        with currentSite(sitemanc):
            assert component.getSiteManager() == sitemanc.getSiteManager()
            try:
                func(*args)
            finally:
                current_mock_db = None
                if teardown:
                    teardown()

    return f

def WithMockDS(*args, **kwargs):
    teardown = lambda: None
    db = ZODB.DB(DemoStorage(name='Users'))
    if len(args) == 1 and not kwargs:
        # Being used as a plain decorator
        func = args[0]
        return _mock_ds_wrapper_for(func, db, teardown)
    return lambda func: _mock_ds_wrapper_for(func, db , teardown)


class SharedConfiguringTestLayer(ZopeComponentLayer,
                                 GCLayerMixin,
                                 ConfiguringLayerMixin):

    set_up_packages = ('nti.site',)

    @classmethod
    def db(cls):
        return current_mock_db

    @classmethod
    def setUp(cls):
        setHooks()
        cls.setUpPackages()
        # Force all the thresholds low so that we do as much testing as possible
        # with btrees.
        from ..site import BTreeLocalAdapterRegistry
        from ..folder import HostPolicySiteManager
        assert hasattr(HostPolicySiteManager, 'btree_threshold')
        HostPolicySiteManager.btree_threshold = 0
        assert hasattr(BTreeLocalAdapterRegistry, 'btree_provided_threshold')
        assert hasattr(BTreeLocalAdapterRegistry, 'btree_map_threshold')
        cls._orig_provided = BTreeLocalAdapterRegistry.btree_provided_threshold
        cls._orig_map = BTreeLocalAdapterRegistry.btree_map_threshold
        BTreeLocalAdapterRegistry.btree_provided_threshold = 0
        BTreeLocalAdapterRegistry.btree_map_threshold = 0

    @classmethod
    def tearDown(cls):
        from ..site import BTreeLocalAdapterRegistry
        from ..folder import HostPolicySiteManager
        del HostPolicySiteManager.btree_threshold
        assert hasattr(HostPolicySiteManager, 'btree_threshold')
        BTreeLocalAdapterRegistry.btree_provided_threshold = cls._orig_provided
        BTreeLocalAdapterRegistry.btree_map_threshold = cls._orig_map
        cls.tearDownPackages()
        zope.testing.cleanup.cleanUp()

    @classmethod
    def testSetUp(cls, test=None):
        setHooks()
        test = test or find_test()
        test.db = cls.db()

    @classmethod
    def testTearDown(cls):
        pass

import unittest

class SiteTestCase(unittest.TestCase):
    layer = SharedConfiguringTestLayer



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Helpers for working with local utilities.

A local utility is a (persistent) utility registered in a local (persistent) site
manager. Local utilities need to be traversable and thus are usually children
of the local site manager itself.

.. $Id$
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

from zope import component
from .utils import registerUtility
from .utils import unregisterUtility

def install_utility(utility, utility_name, provided, local_site_manager):
    """
    Call this to install a local utility. Often you will
    do this from inside a handler for the registration of another
    dependent utility (:class:`.IRegistration`).

    The utility should be :class:`.IContained` because it
    will be held inside the site manager.

    :param str utility_name: The *traversal* name of the utility, not
        the component registration name. This currently only handles
        the default, unnamed registration.

    .. seealso:: :func:`install_utility_on_registration`
    """

    # Contain the utilities we are about to install.
    # Note that for queryNextUtility, etc, to work properly if they
    # use themselves as the context (which seems to be what people do)
    # these need to be children of the SiteManager object: qNU walks from
    # the context to the enclosing site manager, and then looks through ITS
    # bases

    local_site_manager[utility_name] = utility
    registerUtility(local_site_manager, utility, provided=provided)

def install_utility_on_registration(utility, utility_name, provided, event):
    """
    Call this to install a local utility in response to the registration
    of another object.

    The utility should be :class:`IContained` because it
    will be held inside the site manager.

    :param event: The :class:`.IRegistered` event
    """

    registration = event.object
    local_site_manager = registration.registry
    install_utility(utility, utility_name, provided, local_site_manager)

def uninstall_utility_on_unregistration(utility_name, provided, event):
    """
    When a dependent object is unregistered, this undoes the
    work done by :func:`install_utility`.

    :param str utility_name: The *traversal* name of the utility, not
        the component registration name. This currently only handles
        the default, unnamed registration.

    """

    registration = event.object
    local_site_manager = registration.registry

    child_component = local_site_manager[utility_name]

    looked_up = local_site_manager.getUtility(provided)
    assert looked_up is child_component, (looked_up, child_component)

    unregisterUtility(local_site_manager, child_component, provided=provided)
    del local_site_manager[utility_name]

def queryNextUtility(context, interface, default=None):
    """
    Our persistent sites are a mix of persistent and non-persistent
    bases, with many of them having multiple bases. For example (using
    the notation: name (bases,) [type])::

        example.com (GlobalSiteManager) [global]
        Application (GlobalSiteManager) [persistent]
        site-root.example.com (Application, example.com) [persistent]
        child.example.com (example.com) [global]
        site-child.example.com (child.example.com, site-root.example.com) [persistent]

    This gives site-child.example.com this (correct) resolution order::

        site-child.example.com, child.example.com, site-root.example,com Application, example.com, GSM

    However, :func:`zope.component.queryNextUtility` only looks in the *first* base to find
    a next utility. Therefore, when site-janux.ou.edu asks for a next utility,
    instead of getting something persistent from site-platform.ou.edu,
    it instead gets the global non-persistent version from platform.ou.edu.

    We don't generally want to change the resolution order, but we do need
    to tweak it here for getting next utilities so that we consider
    persistent things first. Note that this breaks down if we have
    utilities registered both persistently and non-persistently at the same level.
    """

    try:
        sm = component.getSiteManager(context)
    except LookupError:
        return default

    # These are returned starting from the GlobalSiteManager
    # and working down the resolution chain
    all_utilities = sm.getAllUtilitiesRegisteredFor(interface)
    if not all_utilities:
        return default

    try:
        me = all_utilities.index(context)
        next_ = me - 1
    except ValueError:
        # Not in it. That means our site manager is the global site manager,
        # and we hit the global catalog
        assert sm == component.getGlobalSiteManager()
        next_ = 0

    result = default
    if next_ >= 0:
        result = all_utilities[next_]
        if result is context:
            # in the GSM, we're querying for the GSM utility?
            # Terminate with default to avoid an infinite loop
            result = default # pragma: no cover
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Helpers for running jobs in specific sites.
.. $Id$
"""

from __future__ import print_function, unicode_literals, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

import warnings
import contextlib

from zope import component
from zope import interface

from zope.component.hooks import site as current_site

from ZODB.interfaces import IDatabase

from nti.site.interfaces import SiteNotInstalledError

from nti.site.interfaces import ITransactionSiteNames
from nti.site.interfaces import ISiteTransactionRunner

from nti.site.site import get_site_for_site_names

@contextlib.contextmanager
def _connection_cm():
    """
    Opens a connection to the default database.
    """
    db = component.getUtility(IDatabase)
    conn = db.open()
    for c in conn.connections.values():
        c.setDebugInfo("_connection_cm")
    try:
        yield conn
    finally:
        conn.close()

@contextlib.contextmanager
def _site_cm(conn, site_names=(), root_folder_name='nti.dataserver'):
    # If we don't sync, then we can get stale objects that
    # think they belong to a closed connection
    # TODO: Are we doing something in the wrong order? Connection
    # is an ISynchronizer and registers itself with the transaction manager,
    # so we shouldn't have to do this manually
    # ... I think the problem was a bad site. I think this can go away.
    # conn.sync()
    # In fact, it must go away; if we sync the conn, we lose the
    # current transaction
    sitemanc = conn.root()[root_folder_name]
    # Put into a policy if need be
    sitemanc = get_site_for_site_names(site_names, sitemanc)

    with current_site(sitemanc):
        if component.getSiteManager() != sitemanc.getSiteManager():
            raise SiteNotInstalledError("Hooks not installed?")
        # XXX: Used to do this check...is it really needed?
        # if component.getUtility( interfaces.IDataserver ) is None:
        #   raise InappropriateSiteError()
        yield sitemanc

from nti.transactions.transactions import TransactionLoop

class _RunJobInSite(TransactionLoop):

    def __init__(self, *args, **kwargs):
        self.site_names = kwargs.pop('site_names')
        self.job_name = kwargs.pop('job_name')
        self.side_effect_free = kwargs.pop('side_effect_free')
        self.root_folder_name = kwargs.pop('root_folder_name')
        super(_RunJobInSite, self).__init__(*args, **kwargs)

    def describe_transaction(self, *args, **kwargs):
        if self.job_name:
            return self.job_name
        # Derive from the function
        func = self.handler
        note = func.__doc__
        if note:
            note = note.split('\n', 1)[0]
        else:
            note = func.__name__
        return note

    def run_handler(self, conn, *args, **kwargs):
        with _site_cm(conn, self.site_names, self.root_folder_name):
            for c in conn.connections.values():
                c.setDebugInfo(self.site_names)
            result = self.handler(*args, **kwargs)

            # Commit the transaction while the site is still current
            # so that any before-commit hooks run with that site
            # (Though this has the problem that after-commit hooks would have an invalid
            # site!)
            # JAM: DISABLED because the pyramid requests never ran like this:
            # they commit after they are done and the site has been removed
            # t.commit()

            return result

    def __call__(self, *args, **kwargs):
        with _connection_cm() as conn:
            for c in conn.connections.values():
                c.setDebugInfo(self.describe_transaction(*args, **kwargs))
            # Notice we don't keep conn as an ivar anywhere, to avoid
            # any chance of circular references. These need to be sure to be
            # reclaimed
            return super(_RunJobInSite, self).__call__(conn, *args, **kwargs)

_marker = object()

def get_possible_site_names(*args, **kwargs):
    """
    Helper to find the most applicable site names.

    This uses the :class:`.ITransactionSiteNames` utility.
    """
    utility = component.queryUtility(ITransactionSiteNames)
    result = utility(*args, **kwargs) if utility is not None else None
    return result

@interface.provider(ISiteTransactionRunner)
def run_job_in_site(func,
                    retries=0,
                    sleep=None,
                    site_names=_marker,
                    job_name=None,
                    side_effect_free=False,
                    root_folder_name='nti.dataserver'):
    """
    Runs the function given in `func` in a transaction and dataserver local
    site manager. See :class:`.ISiteTransactionRunner`

    :return: The value returned by the first successful invocation of `func`.
    """

    # site_names is deprecated, we want to start preserving
    # the current site. Because the current site should be based on the
    # current site names FOR NOW, preserving the current site names
    # is equivalent. THIS IS CHANGING though.
    if site_names is not _marker:
        warnings.warn("site_names is deprecated. "
                      "Call this already in the appropriate site",
                      FutureWarning)
    else:
        site_names = get_possible_site_names()

    return _RunJobInSite( func,
                          retries=retries,
                          sleep=sleep,
                          site_names=site_names,
                          job_name=job_name,
                          side_effect_free=side_effect_free,
                          root_folder_name=root_folder_name)()

run_job_in_site.__doc__ = ISiteTransactionRunner['__call__'].getDoc()
